﻿using System;
using System.CodeDom;
using System.CodeDom.Compiler;
using System.Linq;
using System.Net;
using System.Reflection;
using System.ServiceModel;
using System.Text;
using System.Web.Services;
using System.Web.Services.Protocols;
using Microsoft.CSharp;
using System.Web.Services.Description;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Web.Services.Discovery;
using System.Xml.Schema;

namespace Test.Infrastructure
{
    /// <summary>
    /// 
    /// </summary>
    public static class WebServiceClientHelper
    {
        static IDictionary<int, DefaultWebServiceClient> _methodTypeInfo = new Dictionary<int, DefaultWebServiceClient>();

        public static DefaultWebServiceClient GetClient(Stream wsdl, string url, string methodname, string providerName)
        {
            int hsCode = url.GetHashCode();
            DefaultWebServiceClient client;
            lock (_methodTypeInfo)
            {
                if (!_methodTypeInfo.TryGetValue(hsCode, out client))
                {
                    client = new DefaultWebServiceClient(wsdl, url, methodname, providerName);
                    _methodTypeInfo.Add(hsCode, client);
                }
            }
            return client;
        }

        public static object InvokeWebService(string providerName, string url, string methodName, object[] args, WebClient wc = null)
        {
            object result = null;

            if (wc == null) wc = new WebClient();
            using (wc)
            {
                using (Stream wsdl = wc.OpenRead(url))
                {
                    var client = GetClient(wsdl, url, methodName, providerName);
                    client.SetValue("Timeout", wsdl.ReadTimeout);
                    result = client.InvokeService(args);
                }
            }

            return result;
        }
    }

    public class HttpWebClient : WebClient
    {
        /// <summary>
        /// 初始化需要设置超时时间，以毫秒为单位
        /// </summary>
        /// <param name="timeout">毫秒为单位</param>
        public HttpWebClient(int timeout)
        {
            Timeout = timeout;
        }

        public int Timeout { get; set; }

        /// <summary>
        /// 重写 GetWebRequest，添加 WebRequest 对象超时时间
        /// </summary>
        protected override WebRequest GetWebRequest(Uri address)
        {
            HttpWebRequest request = (HttpWebRequest)base.GetWebRequest(address);
            request.Timeout = Timeout;
            request.ReadWriteTimeout = Timeout;
            return request;
        }
    }


    public class DefaultWebServiceClient
    {
        Type _type;
        MethodInfo _method;
        object _obj;

        public object InvokeService(object[] args)
        {
            object proxy = GetProxy();
            return _method.Invoke(proxy, args);
        }

        public void SetValue(string fieldName, object value)
        {
            object proxy = GetProxy();
            PropertyInfo field = _type.GetProperty(fieldName);
            if (field != null)
                field.SetValue(proxy, value);
        }

        public object GetProxy()
        {
            if (_obj == null)
                _obj = Activator.CreateInstance(_type);

            return _obj;
        }

        public MethodInfo MethodInfo
        {
            get { return _method; }
        }

        public DefaultWebServiceClient(Stream wsdl, string url, string methodname, string providerName)
        {
            if (wsdl == null || (wsdl.CanWrite && wsdl.Length == 0))
                throw new Exception("Wsdl为空");

            try
            {
                ServiceDescription sd = ServiceDescription.Read(wsdl);
                ServiceDescriptionImporter sdi = new ServiceDescriptionImporter();
                sdi.AddServiceDescription(sd, "", "");
                CodeNamespace cn = new CodeNamespace(string.Format("DefaultWebServiceClient_{0}_{1}", providerName, wsdl.GetHashCode().ToString()));

                DiscoveryClientProtocol dcp = new DiscoveryClientProtocol();
                dcp.DiscoverAny(url);
                dcp.ResolveAll();
                foreach (object osd in dcp.Documents.Values)
                {
                    if (osd is ServiceDescription) sdi.AddServiceDescription((ServiceDescription)osd, null, null); ;
                    if (osd is XmlSchema) sdi.Schemas.Add((XmlSchema)osd);
                }

                //生成客户端代理类代码 
                CodeCompileUnit ccu = new CodeCompileUnit();
                ccu.Namespaces.Add(cn);
                sdi.Import(cn, ccu);
                CSharpCodeProvider csc = new CSharpCodeProvider();
                ICodeCompiler icc = csc.CreateCompiler();

                //设定编译器的参数 
                CompilerParameters cplist = new CompilerParameters();
                cplist.GenerateExecutable = false;
                cplist.GenerateInMemory = true;
                cplist.ReferencedAssemblies.Add("System.dll");
                cplist.ReferencedAssemblies.Add("System.XML.dll");
                cplist.ReferencedAssemblies.Add("System.Web.Services.dll");
                cplist.ReferencedAssemblies.Add("System.Data.dll");

                //编译代理类 
                CompilerResults cr = icc.CompileAssemblyFromDom(cplist, ccu);
                if (true == cr.Errors.HasErrors)
                {
                    System.Text.StringBuilder sb = new StringBuilder();
                    foreach (CompilerError ce in cr.Errors)
                    {
                        sb.Append(ce.ToString());
                        sb.Append(System.Environment.NewLine);
                    }
                    throw new Exception(sb.ToString());
                }

                //生成代理实例,并调用方法 
                Assembly assembly = cr.CompiledAssembly;
                Type type = null;
                foreach (Type t in assembly.GetExportedTypes())
                {
                    if (t.GetCustomAttributes(typeof(System.Web.Services.WebServiceBindingAttribute), false).Count() > 0)
                    {
                        type = t;
                        break;
                    }
                }

                MethodInfo mi = null;

                if (type == null)
                {
                    throw new Exception("从Wsdl中找不到可调用的类型");
                }

                mi = type.GetMethod(methodname);
                if (mi == null)
                {
                    throw new Exception(string.Format("从Wsdl中找不到可调用的方法{0}.{1}", type.Name, methodname));
                }
                _type = type;
                _method = mi;
            }
            catch (Exception ex)
            {
                throw new Exception("创建WebService客户端失败！", ex);
            }
        }
    }
}
